class Queue<T extends { run: () => Promise<any> }> {
  private start?: Node<T>;
  private status: "waiting" | "running" = "waiting";

  append(item: T, autoRun = true) {
    const node = new Node(item);
    if (this.start) {
      this.start?.setNext(node);
    } else {
      this.start = node;
    }

    if (autoRun) {
      this.startRun();
    }
  }

  startRun() {
    if (this.status === "waiting") {
      this.run();
    }
  }

  async run() {
    if (this.start) {
      this.status = "running";
      const self = this.start.getSelf();
      this.start = this.start.getNext();
      const res = await self.run();
      if (res === false) {
        this.status = "waiting";
      } else {
        this.run();
      }
    } else {
      this.status = "waiting";
    }
  }
}

class Node<T = any> {
  private prev?: Node<T>;
  private next?: Node<T>;
  private self: T;

  constructor(self: T) {
    this.self = self;
  }

  getSelf() {
    return this.self;
  }

  setPrev(prev?: Node<T>) {
    this.prev = prev;
    if (prev) {
      prev.next = this;
    }
  }

  setNext(next?: Node<T>) {
    this.next = next;
    if (next) {
      next.prev = this;
    }
  }

  getPrev() {
    return this.prev;
  }

  getNext() {
    return this.next;
  }
}

export { Queue };
